@@ -46,6 +46,8 @@ G_BEGIN_DECLS
 #define NM_CLIENT_NETWORKING_ENABLED "networking-enabled"
 #define NM_CLIENT_WIRELESS_ENABLED "wireless-enabled"
 #define NM_CLIENT_WIRELESS_HARDWARE_ENABLED "wireless-hardware-enabled"
+#define NM_CLIENT_WWAN_ENABLED "wwan-enabled"
+#define NM_CLIENT_WWAN_HARDWARE_ENABLED "wwan-hardware-enabled"
 #define NM_CLIENT_ACTIVE_CONNECTIONS "active-connections"
 
 typedef struct {
@@ -58,6 +60,14 @@ typedef struct {
 	/* Signals */
 	void (*device_added) (NMClient *client, NMDevice *device);
 	void (*device_removed) (NMClient *client, NMDevice *device);
+
+	/* Padding for future expansion */
+	void (*_reserved1) (void);
+	void (*_reserved2) (void);
+	void (*_reserved3) (void);
+	void (*_reserved4) (void);
+	void (*_reserved5) (void);
+	void (*_reserved6) (void);
 } NMClientClass;
 
 GType nm_client_get_type (void);
@@ -85,6 +95,11 @@ void      nm_client_networking_set_enabled (NMClient *client, gboolean enabled);
 gboolean  nm_client_wireless_get_enabled (NMClient *client);
 void      nm_client_wireless_set_enabled (NMClient *client, gboolean enabled);
 gboolean  nm_client_wireless_hardware_get_enabled (NMClient *client);
+
+gboolean  nm_client_wwan_get_enabled (NMClient *client);
+void      nm_client_wwan_set_enabled (NMClient *client, gboolean enabled);
+gboolean  nm_client_wwan_hardware_get_enabled (NMClient *client);
+
 NMState   nm_client_get_state            (NMClient *client);
 gboolean  nm_client_get_manager_running  (NMClient *client);
 const GPtrArray *nm_client_get_active_connections (NMClient *client);
